import inspect
from typing import Iterable, Sequence, Tuple, TypeVar

from pydantic import BaseModel
from pydantic.fields import (
    MAPPING_LIKE_SHAPES,
    SHAPE_FROZENSET,
    SHAPE_LIST,
    SHAPE_SEQUENCE,
    SHAPE_SET,
    SHAPE_TUPLE,
    SHAPE_TUPLE_ELLIPSIS,
    ModelField,
)

T = TypeVar("T")


def grouped(items: Sequence[T], n: int = 2) -> Iterable[Tuple[T, ...]]:
    """s -> [(s0, s1, s2,...sn-1), (sn, sn+1 , sn+2,...s2n-1), ...]
    list(grouped([1, 2], 2)) == [(1, 2)]
    list(grouped([1, 2, 3, 4], 2)) == [(1, 2), (3, 4)]
    """
    if len(items) % n != 0:
        raise ValueError("items must be equally divisible by n")
    return zip(*[iter(items)] * n)


def is_sequence_like(field: ModelField) -> bool:
    return field.shape in (
        SHAPE_TUPLE,
        SHAPE_TUPLE_ELLIPSIS,
        SHAPE_LIST,
        SHAPE_SET,
        SHAPE_FROZENSET,
        SHAPE_LIST,
        SHAPE_SEQUENCE,
    )


def is_mapping_like(field: ModelField) -> bool:
    return (
        field.shape in MAPPING_LIKE_SHAPES
        or inspect.isclass(field.type_)
        and issubclass(field.type_, BaseModel)
    )



from starlette.exceptions import HTTPException
from starlette.requests import Request
from starlette.responses import JSONResponse

from xpresso.encoders.json import JsonableEncoder
from xpresso.exceptions import RequestValidationError

encoder = JsonableEncoder()


async def http_exception_handler(request: Request, exc: HTTPException) -> JSONResponse:
    headers = getattr(exc, "headers", None)
    if headers:
        return JSONResponse(
            {"detail": exc.detail}, status_code=exc.status_code, headers=headers
        )
    else:
        return JSONResponse({"detail": exc.detail}, status_code=exc.status_code)


async def validation_exception_handler(
    request: Request, exc: RequestValidationError
) -> JSONResponse:
    return JSONResponse(
        encoder({"detail": exc.errors()}),
        status_code=exc.status_code,
    )



import inspect
import sys
from typing import Any, Dict, Optional

if sys.version_info < (3, 8):
    from typing_extensions import Protocol
else:
    from typing import Protocol

from pydantic.schema import TypeModelOrEnum

from xpresso.openapi import models

ModelNameMap = Dict[TypeModelOrEnum, str]
Schemas = Dict[str, Any]


class OpenAPIBody(Protocol):
    def get_openapi(
        self, model_name_map: ModelNameMap, schemas: Schemas
    ) -> models.RequestBody:
        raise NotImplementedError

    def get_schema(
        self, model_name_map: ModelNameMap, schemas: Schemas
    ) -> models.Schema:
        raise NotImplementedError

    def get_media_type(self) -> str:
        raise NotImplementedError

    def get_media_type_object(
        self, model_name_map: ModelNameMap, schemas: Schemas
    ) -> models.MediaType:
        raise NotImplementedError

    def get_encoding(self) -> Optional[models.Encoding]:
        raise NotImplementedError

    def get_field_name(self) -> str:
        raise NotImplementedError


class OpenAPIBodyMarker(Protocol):
    def register_parameter(self, param: inspect.Parameter) -> OpenAPIBody:
        raise NotImplementedError


class OpenAPIParameter(Protocol):
    in_: str
    name: str

    def get_openapi(
        self, model_name_map: ModelNameMap, schemas: Schemas
    ) -> models.ConcreteParameter:
        raise NotImplementedError

    def __eq__(self, o: object) -> bool:
        """Used to de-duplicate parameters declared in multiple places.

        Note that a parameter (defined by in_ and name) should not exist
        more than once with different options.
        This function should take into account _all_ options (e.g. style and explode).
        """
        raise NotImplementedError


class OpenAPIParameterMarker(Protocol):
    def register_parameter(self, param: inspect.Parameter) -> OpenAPIParameter:
        raise NotImplementedError



import typing

from pydantic.error_wrappers import ErrorWrapper
from pydantic.fields import ModelField

from xpresso.exceptions import RequestValidationError
from xpresso.typing import Some


def validate(
    values: typing.Optional[Some[typing.Any]],
    *,
    field: ModelField,
    loc: typing.Tuple[typing.Union[str, int], ...]
) -> typing.Any:
    """Validate after extraction. Should only be used by the top-level body"""
    if values is None:
        if field.required is False:
            return field.get_default()
        else:
            raise RequestValidationError(
                [ErrorWrapper(ValueError("Missing required value"), loc=loc)]
            )
    val, err_or_errors = field.validate(values.value, {}, loc=loc)
    if err_or_errors:
        errors: typing.List[ErrorWrapper]
        if isinstance(err_or_errors, ErrorWrapper):
            errors = [err_or_errors]
        else:
            errors = typing.cast(
                typing.List[ErrorWrapper], err_or_errors
            )  # already a list
        raise RequestValidationError(errors)
    return val



import sys
import typing

if sys.version_info < (3, 8):
    from typing_extensions import Literal
else:
    from typing import Literal

import starlette.applications
import starlette.background
import starlette.datastructures
import starlette.exceptions
import starlette.requests
import starlette.responses
import starlette.routing
import starlette.status
import starlette.types
from di import AsyncExecutor, BaseContainer, ConcurrentAsyncExecutor
from di.api.dependencies import DependantBase
from di.api.executor import AsyncExecutorProtocol
from di.api.providers import DependencyProvider as Endpoint
from di.api.solved import SolvedDependant
from di.dependant import JoinedDependant
from starlette.routing import Host as Host  # noqa: F401
from starlette.routing import Mount as Mount  # noqa: F401

import xpresso._utils.asgi_scope_extension as asgi_scope_extension
import xpresso.binders.dependants as param_dependants
import xpresso.openapi.models as openapi_models
from xpresso.dependencies.models import Dependant
from xpresso.encoders.api import Encoder
from xpresso.encoders.json import JsonableEncoder
from xpresso.exceptions import HTTPException
from xpresso.responses import ResponseSpec

__all__ = (
    "Router",
    "Operation",
    "Path",
    "Mount",
    "Host",
)

Method = Literal[
    "GET", "HEAD", "POST", "PUT", "PATCH", "DELETE", "CONNECT", "OPTIONS", "TACE"
]


class _OperationApp:
    __slots__ = ("dependant", "executor", "response_factory", "response_encoder")

    def __init__(
        self,
        dependant: SolvedDependant[typing.Any],
        executor: AsyncExecutorProtocol,
        response_factory: typing.Callable[[typing.Any], starlette.responses.Response],
        response_encoder: Encoder,
    ) -> None:
        self.dependant = dependant
        self.executor = executor
        self.response_factory = response_factory
        self.response_encoder = response_encoder

    async def __call__(
        self,
        scope: starlette.types.Scope,
        receive: starlette.types.Receive,
        send: starlette.types.Send,
    ) -> None:
        request = starlette.requests.Request(scope=scope, receive=receive, send=send)
        values: typing.Dict[typing.Any, typing.Any] = {
            starlette.requests.Request: request,
            starlette.requests.HTTPConnection: request,
            starlette.background.BackgroundTasks: starlette.background.BackgroundTasks(),
        }
        xpresso_scope: asgi_scope_extension.xpressoASGIExtension = scope["extensions"][
            "xpresso"
        ]
        async with xpresso_scope["container"].enter_scope("operation") as container:
            endpoint_return = await container.execute_async(
                self.dependant,
                values=values,
                executor=self.executor,
            )
            if isinstance(endpoint_return, starlette.responses.Response):
                response = endpoint_return
            else:
                response = self.response_factory(self.response_encoder(endpoint_return))
            xpresso_scope["response"] = response
        await xpresso_scope["response"](scope, receive, send)


class Operation(starlette.routing.BaseRoute):
    def __init__(
        self,
        endpoint: Endpoint,
        *,
        # OpenAPI parameters
        include_in_schema: bool = True,
        tags: typing.Optional[typing.Sequence[str]] = None,
        summary: typing.Optional[str] = None,
        description: typing.Optional[str] = None,
        deprecated: typing.Optional[bool] = None,
        operation_id: typing.Optional[bool] = None,
        servers: typing.Optional[typing.Sequence[openapi_models.Server]] = None,
        external_docs: typing.Optional[openapi_models.ExternalDocumentation] = None,
        responses: typing.Optional[
            typing.Mapping[
                typing.Union[int, str],
                typing.Union[ResponseSpec, typing.Iterable[ResponseSpec]],
            ]
        ] = None,
        # xpresso params
        name: typing.Optional[str] = None,
        dependencies: typing.Optional[typing.Sequence[Dependant]] = None,
        execute_dependencies_concurrently: bool = False,
        response_factory: typing.Callable[
            [typing.Any], starlette.responses.Response
        ] = starlette.responses.JSONResponse,
        response_encoder: Encoder = JsonableEncoder(),
        sync_to_thread: bool = True,
    ) -> None:
        self._app: typing.Optional[starlette.types.ASGIApp] = None
        self.endpoint = endpoint
        self.tags = tags
        self.summary = summary
        self.description = description
        self.deprecated = deprecated
        self.operation_id = operation_id
        self.servers = servers
        self.external_docs = external_docs
        self.responses = dict(responses or {})
        self.dependencies = dependencies or []
        self.execute_dependencies_concurrently = execute_dependencies_concurrently
        self.response_factory = response_factory
        self.response_encoder = response_encoder
        self.include_in_schema = include_in_schema
        self.name: str = starlette.routing.get_name(endpoint) if name is None else name  # type: ignore
        self.sync_to_thread = sync_to_thread

    async def handle(
        self,
        scope: starlette.types.Scope,
        receive: starlette.types.Receive,
        send: starlette.types.Send,
    ) -> None:
        if self._app is None:
            raise RuntimeError("Operation cannot be used outside of a xpresso router")
        return await self._app(scope, receive, send)

    def solve(
        self,
        container: BaseContainer,
        dependencies: typing.List[DependantBase[typing.Any]],
    ) -> None:
        self.dependant = container.solve(
            JoinedDependant(
                Dependant(self.endpoint, sync_to_thread=self.sync_to_thread),
                siblings=[*dependencies, *(self.dependencies or ())],
            )
        )
        flat = self.dependant.get_flat_subdependants()
        bodies = [dep for dep in flat if isinstance(dep, param_dependants.BodyBinder)]
        if len(bodies) > 1:
            raise TypeError("There can only be 1 top level body per operation")
        executor: AsyncExecutorProtocol
        if self.execute_dependencies_concurrently:
            executor = ConcurrentAsyncExecutor()
        else:
            executor = AsyncExecutor()
        self._app = _OperationApp(
            dependant=self.dependant,
            executor=executor,
            response_encoder=self.response_encoder,
            response_factory=self.response_factory,
        )

    def url_path_for(
        self, name: str, **path_params: str
    ) -> starlette.datastructures.URLPath:
        if path_params:
            raise TypeError("Operations cannot have path parameters")
        if name != self.name:
            raise starlette.routing.NoMatchFound()
        return starlette.datastructures.URLPath("/")

    def __eq__(self, __o: object) -> bool:
        return isinstance(__o, type(self)) and self.endpoint is __o.endpoint


class _PathApp:
    """Thin class wrapper so that Starlette treats us as an ASGI App"""

    __slots__ = ("operations",)

    def __init__(self, operations: typing.Dict[str, Operation]) -> None:
        self.operations = operations

    def __call__(
        self,
        scope: starlette.types.Scope,
        receive: starlette.types.Receive,
        send: starlette.types.Send,
    ) -> typing.Awaitable[None]:
        if scope["method"] not in self.operations:
            raise HTTPException(status_code=405, detail="Method not allowed")
        return self.operations[scope["method"]].handle(scope, receive, send)


class Path(starlette.routing.Route):
    def __init__(
        self,
        path: str,
        *,
        get: typing.Optional[typing.Union[Operation, Endpoint]] = None,
        head: typing.Optional[typing.Union[Operation, Endpoint]] = None,
        post: typing.Optional[typing.Union[Operation, Endpoint]] = None,
        put: typing.Optional[typing.Union[Operation, Endpoint]] = None,
        patch: typing.Optional[typing.Union[Operation, Endpoint]] = None,
        delete: typing.Optional[typing.Union[Operation, Endpoint]] = None,
        connect: typing.Optional[typing.Union[Operation, Endpoint]] = None,
        options: typing.Optional[typing.Union[Operation, Endpoint]] = None,
        trace: typing.Optional[typing.Union[Operation, Endpoint]] = None,
        redirect_slashes: bool = True,
        dependencies: typing.Optional[typing.Sequence[Dependant]] = None,
        # OpenAPI metadata
        include_in_schema: bool = True,
        name: typing.Optional[str] = None,
        summary: typing.Optional[str] = None,
        description: typing.Optional[str] = None,
        servers: typing.Optional[typing.Sequence[openapi_models.Server]] = None,
        parameters: typing.Optional[
            typing.Sequence[param_dependants.ParameterBinderMarker]
        ] = None,
    ) -> None:
        if not path.startswith("/"):
            raise ValueError("Routed paths must start with '/'")
        self.path = path
        self.redirect_slashes = redirect_slashes
        self.dependencies = dependencies or []
        self.summary = summary
        self.description = description
        self.servers = servers
        self.parameters = parameters

        operations: typing.Dict[str, Operation] = {}
        if get:
            operations["GET"] = get if isinstance(get, Operation) else Operation(get)
        if head:
            operations["HEAD"] = (
                head if isinstance(head, Operation) else Operation(head)
            )
        if post:
            operations["POST"] = (
                post if isinstance(post, Operation) else Operation(post)
            )
        if put:
            operations["PUT"] = put if isinstance(put, Operation) else Operation(put)
        if patch:
            operations["PATCH"] = (
                patch if isinstance(patch, Operation) else Operation(patch)
            )
        if delete:
            operations["DELETE"] = (
                delete if isinstance(delete, Operation) else Operation(delete)
            )
        if connect:
            operations["CONNECT"] = (
                connect if isinstance(connect, Operation) else Operation(connect)
            )
        if options:
            operations["OPTIONS"] = (
                options if isinstance(options, Operation) else Operation(options)
            )
        if trace:
            operations["TRACE"] = (
                trace if isinstance(trace, Operation) else Operation(trace)
            )
        self.operations = operations
        super().__init__(  # type: ignore  # for Pylance
            path=path,
            endpoint=_PathApp(operations),
            name=name,  # type: ignore[arg-type]
            include_in_schema=include_in_schema,
            methods=list(operations.keys()),
        )


def _not_supported(method: str) -> typing.Callable[..., typing.Any]:
    def raise_error(*args: typing.Any, **kwargs: typing.Any) -> typing.NoReturn:
        raise NotImplementedError(
            f"Use of Router.{method} is deprecated."
            " Use Router(routes=[...]) instead."
        )

    return raise_error


class Router(starlette.routing.Router):
    routes: typing.List[starlette.routing.BaseRoute]

    def __init__(
        self,
        routes: typing.Sequence[starlette.routing.BaseRoute],
        *,
        redirect_slashes: bool = True,
        default: typing.Optional[starlette.types.ASGIApp] = None,
        lifespan: typing.Optional[
            typing.Callable[
                [starlette.applications.Starlette], typing.AsyncContextManager[None]
            ]
        ] = None,
        dependencies: typing.Optional[typing.List[Dependant]] = None,
    ) -> None:
        self.dependencies = dependencies or []
        super().__init__(  # type: ignore
            routes=list(routes),
            redirect_slashes=redirect_slashes,
            default=default,  # type: ignore
            lifespan=lifespan,  # type: ignore
        )

    mount = _not_supported("mount")
    host = _not_supported("host")
    add_route = _not_supported("add_route")
    add_websocket_route = _not_supported("add_websocket_route")
    route = _not_supported("route")
    websocket_route = _not_supported("websocket_route")
    add_event_handler = _not_supported("add_event_handler")
    on_event = _not_supported("on_event")



from typing import Any, Awaitable, Callable

from xpresso.openapi.models import SecurityBase as SecurityBaseModel


class SecurityBase:
    model: SecurityBaseModel
    scheme_name: str
    __call__: Callable[..., Awaitable[Any]]



from typing import Optional

from starlette.exceptions import HTTPException
from starlette.requests import Request
from starlette.status import HTTP_401_UNAUTHORIZED

from xpresso.openapi.models import APIKey
from xpresso.security._base import SecurityBase


class APIKeyBase(SecurityBase):
    pass


class APIKeyQuery(APIKeyBase):
    def __init__(
        self,
        *,
        name: str,
        scheme_name: Optional[str] = None,
        description: Optional[str] = None,
        auto_error: bool = True
    ):
        self.model: APIKey = APIKey(
            **{"in": "query"}, name=name, description=description
        )
        self.scheme_name = scheme_name or self.__class__.__name__
        self.auto_error = auto_error

    async def __call__(self, request: Request) -> Optional[str]:
        api_key: str = request.query_params.get(self.model.name)
        if not api_key:
            if self.auto_error:
                raise HTTPException(
                    status_code=HTTP_401_UNAUTHORIZED, detail="Not authenticated"
                )
            else:
                return None
        return api_key


class APIKeyHeader(APIKeyBase):
    def __init__(
        self,
        *,
        name: str,
        scheme_name: Optional[str] = None,
        description: Optional[str] = None,
        auto_error: bool = True
    ):
        self.model: APIKey = APIKey(
            **{"in": "header"}, name=name, description=description
        )
        self.scheme_name = scheme_name or self.__class__.__name__
        self.auto_error = auto_error

    async def __call__(self, request: Request) -> Optional[str]:
        api_key: str = request.headers.get(self.model.name)
        if not api_key:
            if self.auto_error:
                raise HTTPException(
                    status_code=HTTP_401_UNAUTHORIZED, detail="Not authenticated"
                )
            else:
                return None
        return api_key


class APIKeyCookie(APIKeyBase):
    def __init__(
        self,
        *,
        name: str,
        scheme_name: Optional[str] = None,
        description: Optional[str] = None,
        auto_error: bool = True
    ):
        self.model: APIKey = APIKey(
            **{"in": "cookie"}, name=name, description=description
        )
        self.scheme_name = scheme_name or self.__class__.__name__
        self.auto_error = auto_error

    async def __call__(self, request: Request) -> Optional[str]:
        api_key = request.cookies.get(self.model.name)
        if not api_key:
            if self.auto_error:
                raise HTTPException(
                    status_code=HTTP_401_UNAUTHORIZED, detail="Not authenticated"
                )
            else:
                return None
        return api_key



import typing
from dataclasses import dataclass

from starlette.routing import BaseRoute, Mount, Router


@dataclass(frozen=True)
class VisitedRoute:
    path: str
    routers: typing.List[Router]
    route: BaseRoute


def visit_routes(
    routers: typing.List[Router], path: typing.Optional[str] = None
) -> typing.Generator[VisitedRoute, None, None]:
    path = path or ""
    router = next(iter(reversed(routers)), None)
    assert router is not None
    for route in typing.cast(typing.Iterable[BaseRoute], router.routes):
        if isinstance(route, Mount) and isinstance(route.app, Router):
            yield from visit_routes(routers + [route.app], path + route.path)
        elif hasattr(route, "path"):
            yield VisitedRoute(path + getattr(route, "path"), routers, route)



import functools
import inspect
from dataclasses import dataclass
from typing import Any, Callable, ClassVar, Dict, Iterable, List, Optional, Tuple, cast

from pydantic.error_wrappers import ErrorWrapper
from pydantic.fields import ModelField
from starlette.requests import HTTPConnection

from xpresso.binders._extractors.api import ParameterExtractor, ParameterExtractorMarker
from xpresso.binders._extractors.exceptions import InvalidSerialization
from xpresso.binders._extractors.params.base import (
    ParameterExtractorBase,
    get_basic_param_info,
)
from xpresso.binders._extractors.utils import grouped, is_mapping_like, is_sequence_like
from xpresso.exceptions import RequestValidationError
from xpresso.typing import Some


def collect_scalar(name: str, style: str, explode: bool, value: str) -> str:
    if style == "simple":
        return value
    if style == "label":
        if not value.startswith("."):
            raise InvalidSerialization("label serialized parameter must start with '.'")
        return value[1:]
    # style == "matrix"
    template = f";{name}="
    if not value.startswith(template):
        raise InvalidSerialization(
            f"matrix serialized parameter must start with {template}"
        )
    return value[len(template) :]


def collect_sequence(name: str, style: str, explode: bool, value: str) -> List[str]:
    if style == "simple":
        return value.split(",")
    if style == "label":
        if value[0] != ".":
            raise InvalidSerialization("label serialized parameter must start with '.'")
        if explode:
            return value[1:].split(".")
        return value[1:].split(",")
    # style == "matrix"
    template = f";{name}="
    if not value.startswith(template):
        raise InvalidSerialization(
            f"matrix serialized parameter must start with {template}"
        )
    if explode:
        return value.split(template)[1:]
    return value.split(template, maxsplit=1)[1].split(",")


def collect_object(
    name: str,
    style: str,
    explode: bool,
    value: str,
) -> Dict[str, str]:
    if style == "simple":
        delimiter = ","
    elif style == "label":
        value = value[1:]
        if explode:
            delimiter = "."
        else:
            delimiter = ","
    else:  # style == "matrix"
        if explode:
            if value[0] != ";":
                raise InvalidSerialization(
                    "object-valued path parameter could be deserialized"
                    f" with style=matrix, explode=True: {value}"
                )
            value = value[1:]
            delimiter = ";"
        else:
            value = value[len(f";{name}=") :]
            delimiter = ","
    if explode:
        res: Dict[str, str] = {}
        for field in value.split(delimiter):
            split = field.split("=", maxsplit=1)
            if len(split) != 2:
                raise InvalidSerialization(f"{field} is not a valid field encoding")
            name, val = field.split("=", maxsplit=1)
            res[name] = val
        return res
    return dict(cast(Iterable[Tuple[str, str]], grouped(value.split(delimiter))))


def get_extractor(style: str, explode: bool, field: ModelField) -> Callable[..., Any]:
    # form style
    if is_sequence_like(field):
        return functools.partial(
            collect_sequence,
            style=style,
            explode=explode,
        )
    if is_mapping_like(field):
        return functools.partial(
            collect_object,
            style=style,
            explode=explode,
        )
    # single item
    return functools.partial(
        collect_scalar,
        style=style,
        explode=explode,
    )


@dataclass(frozen=True)
class PathParameterExtractor(ParameterExtractorBase):
    extractor: Callable[..., Any]
    in_: ClassVar[str] = "path"

    def extract(self, connection: HTTPConnection) -> Any:
        param_value: str = connection.path_params[self.name]  # type: ignore[assignment]
        try:
            extracted = self.extractor(name=self.name, value=param_value)
        except InvalidSerialization as exc:
            raise RequestValidationError(
                [ErrorWrapper(exc=exc, loc=("path", self.name))]
            )
        return self.validate(Some(extracted))


@dataclass(frozen=True)
class PathParameterExtractorMarker(ParameterExtractorMarker):
    alias: Optional[str]
    explode: bool
    style: str
    in_: ClassVar[str] = "path"

    def register_parameter(self, param: inspect.Parameter) -> ParameterExtractor:
        field, name, loc = get_basic_param_info(param, self.alias, self.in_)
        extractor = get_extractor(style=self.style, explode=self.explode, field=field)
        return PathParameterExtractor(
            field=field, loc=loc, name=name, extractor=extractor
        )



import inspect
import sys
from typing import Any, Iterable, Optional, Union

if sys.version_info < (3, 8):
    from typing_extensions import Protocol
else:
    from typing import Protocol

from starlette.datastructures import FormData, UploadFile
from starlette.requests import HTTPConnection, Request

from xpresso.typing import Some


class BodyExtractor(Protocol):
    async def extract_from_request(self, request: Request) -> Any:
        """Extract from top level request"""
        raise NotImplementedError

    async def extract_from_form(
        self,
        form: FormData,
        *,
        loc: Iterable[Union[str, int]],
    ) -> Optional[Some[Any]]:
        """Extract from a form"""
        raise NotImplementedError

    async def extract_from_field(
        self,
        field: Union[str, UploadFile],
        *,
        loc: Iterable[Union[str, int]],
    ) -> Any:
        """Extract from a form field"""
        raise NotImplementedError

    def matches_media_type(self, media_type: Optional[str]) -> bool:
        """Check if this extractor can extract the given media type"""
        raise NotImplementedError


class BodyExtractorMarker(Protocol):
    def register_parameter(self, param: inspect.Parameter) -> BodyExtractor:
        """Hook to pull information from the Python parameter/field this body is attached to"""
        raise NotImplementedError


class ParameterExtractor(Protocol):
    def extract(self, connection: HTTPConnection) -> Any:
        raise NotImplementedError


class ParameterExtractorMarker(Protocol):
    def register_parameter(self, param: inspect.Parameter) -> ParameterExtractor:
        """Hook to pull information from the Python parameter/field this body is attached to"""
        raise NotImplementedError



import json
import sys
import typing

if sys.version_info < (3, 9):
    from typing_extensions import Annotated
else:
    from typing import Annotated

if sys.version_info < (3, 8):
    from typing_extensions import Literal
else:
    from typing import Literal

import xpresso.binders.dependants as param_dependants
from xpresso.binders._extractors.body.discriminated import (
    ContentTypeDiscriminatedExtractorMarker,
)
from xpresso.binders._extractors.body.field import (
    FieldExtractorMarker,
    RepeatedFieldExtractorMarker,
)
from xpresso.binders._extractors.body.file import FileBodyExtractorMarker
from xpresso.binders._extractors.body.form import (
    FormDataBodyExtractorMarker,
    MultipartBodyExtractorMarker,
)
from xpresso.binders._extractors.body.form_field import FormFieldBodyExtractorMarker
from xpresso.binders._extractors.body.json import Decoder, JsonBodyExtractorMarker
from xpresso.binders._extractors.params.cookie import CookieParameterExtractorMarker
from xpresso.binders._extractors.params.header import HeaderParameterExtractorMarker
from xpresso.binders._extractors.params.path import PathParameterExtractorMarker
from xpresso.binders._extractors.params.query import QueryParameterExtractorMarker
from xpresso.binders._openapi_providers.body.discriminated import (
    OpenAPIContentTypeDiscriminatedMarker,
)
from xpresso.binders._openapi_providers.body.field import (
    OpenAPIFieldMarker,
    OpenAPIRepeatedFieldMarker,
)
from xpresso.binders._openapi_providers.body.file import OpenAPIFileMarker
from xpresso.binders._openapi_providers.body.form import OpenAPIFormDataMarker
from xpresso.binders._openapi_providers.body.form_field import OpenAPIFormFieldMarker
from xpresso.binders._openapi_providers.body.json import OpenAPIJsonMarker
from xpresso.binders._openapi_providers.params.cookie import (
    OpenAPICookieParameterMarker,
)
from xpresso.binders._openapi_providers.params.header import (
    OpenAPIHeaderParameterMarker,
)
from xpresso.binders._openapi_providers.params.path import OpenAPIPathParameterMarker
from xpresso.binders._openapi_providers.params.query import OpenAPIQueryParameterMarker
from xpresso.openapi import models as openapi_models

T = typing.TypeVar("T")

Example = typing.Union[openapi_models.Example, typing.Any]


def QueryParam(
    *,
    alias: typing.Optional[str] = None,
    style: openapi_models.QueryParamStyles = "form",
    explode: bool = True,
    examples: typing.Optional[typing.Dict[str, Example]] = None,
    description: typing.Optional[str] = None,
    deprecated: typing.Optional[bool] = None,
) -> param_dependants.ParameterBinderMarker:
    extractor = QueryParameterExtractorMarker(
        alias=alias,
        explode=explode,
        style=style,
    )
    openapi = OpenAPIQueryParameterMarker(
        alias=alias,
        description=description,
        style=style,
        explode=explode,
        examples=examples,
        deprecated=deprecated,
    )
    return param_dependants.ParameterBinderMarker(
        in_="query",
        extractor_marker=extractor,
        openapi_marker=openapi,
    )


def PathParam(
    *,
    alias: typing.Optional[str] = None,
    style: openapi_models.PathParamStyles = "simple",
    explode: bool = False,
    examples: typing.Optional[typing.Dict[str, Example]] = None,
    description: typing.Optional[str] = None,
    deprecated: typing.Optional[bool] = None,
) -> param_dependants.ParameterBinderMarker:
    extractor = PathParameterExtractorMarker(
        alias=alias,
        explode=explode,
        style=style,
    )
    openapi = OpenAPIPathParameterMarker(
        alias=alias,
        description=description,
        style=style,
        explode=explode,
        examples=examples,
        deprecated=deprecated,
    )
    return param_dependants.ParameterBinderMarker(
        in_="path",
        extractor_marker=extractor,
        openapi_marker=openapi,
    )


def HeaderParam(
    *,
    convert_underscores: bool = True,
    alias: typing.Optional[str] = None,
    explode: bool = False,
    examples: typing.Optional[typing.Dict[str, Example]] = None,
    description: typing.Optional[str] = None,
    deprecated: typing.Optional[bool] = None,
) -> param_dependants.ParameterBinderMarker:
    extractor = HeaderParameterExtractorMarker(
        alias=alias,
        explode=explode,
        convert_underscores=convert_underscores,
    )
    openapi = OpenAPIHeaderParameterMarker(
        alias=alias,
        description=description,
        explode=explode,
        style="simple",
        examples=examples,
        deprecated=deprecated,
    )
    return param_dependants.ParameterBinderMarker(
        in_="header",
        extractor_marker=extractor,
        openapi_marker=openapi,
    )


def CookieParam(
    *,
    alias: typing.Optional[str] = None,
    explode: bool = True,
    examples: typing.Optional[typing.Dict[str, Example]] = None,
    description: typing.Optional[str] = None,
    deprecated: typing.Optional[bool] = None,
) -> param_dependants.ParameterBinderMarker:
    extractor = CookieParameterExtractorMarker(
        alias=alias,
        explode=explode,
    )
    openapi = OpenAPICookieParameterMarker(
        alias=alias,
        description=description,
        style="form",
        explode=explode,
        examples=examples,
        deprecated=deprecated,
    )
    return param_dependants.ParameterBinderMarker(
        in_="cookie",
        extractor_marker=extractor,
        openapi_marker=openapi,
    )


def Json(
    *,
    examples: typing.Optional[typing.Dict[str, Example]] = None,
    description: typing.Optional[str] = None,
    decoder: Decoder = json.loads,
    enforce_media_type: bool = True,
) -> param_dependants.BodyBinderMarker:
    extractor = JsonBodyExtractorMarker(
        decoder=decoder,
        enforce_media_type=enforce_media_type,
    )
    openapi = OpenAPIJsonMarker(
        description=description,
        examples=examples,
    )
    return param_dependants.BodyBinderMarker(
        extractor_marker=extractor,
        openapi_marker=openapi,
    )


def File(
    *,
    media_type: typing.Optional[str] = None,
    enforce_media_type: bool = True,
    examples: typing.Optional[typing.Dict[str, Example]] = None,
    description: typing.Optional[str] = None,
    format: Literal["binary", "base64"] = "binary",
) -> param_dependants.BodyBinderMarker:
    extractor = FileBodyExtractorMarker(
        media_type=media_type,
        enforce_media_type=enforce_media_type,
    )
    openapi = OpenAPIFileMarker(
        description=description,
        examples=examples,
        media_type=media_type,
        format=format,
    )
    return param_dependants.BodyBinderMarker(
        extractor_marker=extractor,
        openapi_marker=openapi,
    )


def FormEncodedField(
    *,
    alias: typing.Optional[str] = None,
    style: openapi_models.FormDataStyles = "form",
    explode: bool = True,
) -> param_dependants.BodyBinderMarker:
    extractor = FormFieldBodyExtractorMarker(
        alias=alias,
        style=style,
        explode=explode,
    )
    openapi = OpenAPIFormFieldMarker(
        alias=alias,
        style=style,
        explode=explode,
    )
    return param_dependants.BodyBinderMarker(
        extractor_marker=extractor,
        openapi_marker=openapi,
    )


def FormField(
    *,
    alias: typing.Optional[str] = None,
) -> param_dependants.BodyBinderMarker:
    extractor = FieldExtractorMarker(alias=alias)
    openapi = OpenAPIFieldMarker(alias=alias)
    return param_dependants.BodyBinderMarker(
        extractor_marker=extractor,
        openapi_marker=openapi,
    )


def RepeatedFormField(
    *,
    alias: typing.Optional[str] = None,
) -> param_dependants.BodyBinderMarker:
    extractor = RepeatedFieldExtractorMarker(alias=alias)
    openapi = OpenAPIRepeatedFieldMarker(alias=alias)
    return param_dependants.BodyBinderMarker(
        extractor_marker=extractor,
        openapi_marker=openapi,
    )


def Form(
    *,
    enforce_media_type: bool = True,
    examples: typing.Optional[typing.Dict[str, Example]] = None,
    description: typing.Optional[str] = None,
) -> param_dependants.BodyBinderMarker:
    extractor = FormDataBodyExtractorMarker(
        enforce_media_type=enforce_media_type,
    )
    openapi = OpenAPIFormDataMarker(
        description=description,
        examples=examples,
        media_type="application/x-www-form-urlencoded",
    )
    return param_dependants.BodyBinderMarker(
        extractor_marker=extractor,
        openapi_marker=openapi,
    )


def Multipart(
    *,
    enforce_media_type: bool = True,
    examples: typing.Optional[typing.Dict[str, Example]] = None,
    description: typing.Optional[str] = None,
) -> param_dependants.BodyBinderMarker:
    extractor = MultipartBodyExtractorMarker(
        enforce_media_type=enforce_media_type,
    )
    openapi = OpenAPIFormDataMarker(
        description=description,
        examples=examples,
        media_type="multipart/form-data",
    )
    return param_dependants.BodyBinderMarker(
        extractor_marker=extractor,
        openapi_marker=openapi,
    )


def ContentTypeDiscriminatedBody(
    *,
    description: typing.Optional[str] = None,
) -> param_dependants.BodyBinderMarker:
    extractor = ContentTypeDiscriminatedExtractorMarker()
    openapi = OpenAPIContentTypeDiscriminatedMarker(
        description=description,
    )
    return param_dependants.BodyBinderMarker(
        extractor_marker=extractor,
        openapi_marker=openapi,
    )


# Convenience type aliases
FromQuery = Annotated[T, QueryParam()]
FromHeader = Annotated[T, HeaderParam()]
FromCookie = Annotated[T, CookieParam()]
FromPath = Annotated[T, PathParam()]
FromJson = Annotated[T, Json()]
FromFile = Annotated[T, File()]
FromFormField = Annotated[T, FormEncodedField()]
ExtractField = Annotated[T, FormField()]
ExtractRepeatedField = Annotated[T, RepeatedFormField()]
FromFormData = Annotated[T, Form()]
FromMultipart = Annotated[T, Multipart()]
ByContentType = Annotated[T, ContentTypeDiscriminatedBody()]



import inspect
import typing
from dataclasses import dataclass
from urllib.parse import unquote_plus

from di.typing import get_markers_from_parameter
from pydantic.fields import ModelField
from starlette.datastructures import FormData, Headers, UploadFile
from starlette.formparsers import FormParser
from starlette.requests import Request

from xpresso._utils.media_type_validator import MediaTypeValidator
from xpresso._utils.media_type_validator import (
    get_validator as get_media_type_validator,
)
from xpresso._utils.typing import model_field_from_param
from xpresso.binders._extractors.api import BodyExtractor, BodyExtractorMarker
from xpresso.binders._extractors.body.form_field import FormFieldBodyExtractorMarker
from xpresso.binders._extractors.validator import validate as validate_data
from xpresso.binders.dependants import BodyBinderMarker
from xpresso.typing import Some


@dataclass(frozen=True)
class FormDataBodyExtractorBase(BodyExtractor):
    media_type_validator: MediaTypeValidator
    field: ModelField
    field_extractors: typing.Dict[str, BodyExtractor]

    def matches_media_type(self, media_type: typing.Optional[str]) -> bool:
        return self.media_type_validator.matches(media_type)

    async def extract_from_request(self, request: Request) -> typing.Any:
        media_type = request.headers.get("content-type", None)
        self.media_type_validator.validate(media_type, loc=("body",))
        if (
            request.headers.get("Content-Length", None) == "0"
            and self.field.required is not True
        ):
            # this is the only way to know the body is empty
            return validate_data(
                None,
                field=self.field,
                loc=("body",),
            )
        return validate_data(
            Some(await self._extract(await request.form(), loc=("body",))),
            field=self.field,
            loc=("body",),
        )

    async def _extract(
        self, form: FormData, loc: typing.Iterable[typing.Union[str, int]]
    ) -> typing.Any:
        res: typing.Dict[str, typing.Any] = {}
        for param_name, extractor in self.field_extractors.items():
            extracted = await extractor.extract_from_form(form, loc=loc)
            if isinstance(extracted, Some):
                res[param_name] = extracted.value
        return res

    # These are implemented to work around pecularities of hashing bound methods on Python 3.7
    def __hash__(self) -> int:
        return id(self)

    def __eq__(self, __o: object) -> bool:
        return self is __o


class FormDataBodyExtractor(FormDataBodyExtractorBase):
    async def extract_from_field(
        self,
        field: typing.Union[str, UploadFile],
        *,
        loc: typing.Iterable[typing.Union[str, int]],
    ) -> typing.Optional[Some[typing.Any]]:
        if isinstance(field, UploadFile):

            async def stream_gen() -> typing.AsyncGenerator[bytes, None]:
                yield await field.read()  # type: ignore # UploadFile always returns bytes
                yield b""

            stream = stream_gen()

            form = await FormParser(Headers(), stream=stream).parse()

        else:
            forms = field.split("&")
            items = typing.cast(
                typing.List[typing.Tuple[str, str]],
                [tuple(unquote_plus(f).split("=", maxsplit=1)) for f in forms],
            )
            form = FormData(items)  # type: ignore[arg-type]

        return await self._extract(form, loc=loc)


class MultipartBodyExtractor(FormDataBodyExtractorBase):
    pass


@dataclass(frozen=True)
class FormDataBodyExtractorMarkerBase(BodyExtractorMarker):
    enforce_media_type: bool
    media_type: typing.ClassVar[str]
    cls: typing.ClassVar[
        typing.Union[
            typing.Type[FormDataBodyExtractor], typing.Type[MultipartBodyExtractor]
        ]
    ]

    def register_parameter(self, param: inspect.Parameter) -> BodyExtractor:
        formdata_field = model_field_from_param(param)

        field_extractors: typing.Dict[str, BodyExtractor] = {}
        # use pydantic to get rid of outer annotated, optional, etc.
        annotation = formdata_field.type_
        for param_name, field_param in inspect.signature(annotation).parameters.items():
            marker: typing.Optional[BodyBinderMarker] = None
            for param_marker in get_markers_from_parameter(field_param):
                if isinstance(param_marker, BodyBinderMarker):
                    marker = param_marker
                    break
            extractor_marker: BodyExtractorMarker
            if marker is None:
                # use the defaults
                extractor_marker = FormFieldBodyExtractorMarker(
                    alias=None,
                    style="form",
                    explode=True,
                )
            else:
                extractor_marker = marker.extractor_marker
            extractor = extractor_marker.register_parameter(field_param)
            field_extractors[param_name] = extractor
        if self.enforce_media_type and self.media_type:
            media_type_validator = get_media_type_validator(self.media_type)
        else:
            media_type_validator = get_media_type_validator(None)
        return self.cls(
            media_type_validator=media_type_validator,
            field_extractors=field_extractors,
            field=formdata_field,
        )


@dataclass(frozen=True)
class FormDataBodyExtractorMarker(FormDataBodyExtractorMarkerBase):
    media_type = "application/x-www-form-urlencoded"
    cls = FormDataBodyExtractor


@dataclass(frozen=True)
class MultipartBodyExtractorMarker(FormDataBodyExtractorMarkerBase):
    media_type = "multipart/form-data"
    cls = MultipartBodyExtractor



from typing import Any, Mapping, Optional, Union

from pydantic import BaseModel, Field
from starlette.requests import Request
from starlette.responses import Response

from xpresso.openapi.models import Example, Header


def get_response(request: Request) -> Response:
    try:
        return request.scope["extensions"]["xpresso"]["response"]  # type: ignore[return-type]
    except KeyError:
        raise LookupError(
            "xpresso.responses.get_response was called"
            " before the endpoint has finished executing or the endpoint raised an exception"
        )


def set_response(request: Request, response: Response) -> None:
    request.scope["extensions"]["xpresso"]["response"] = response  # type: ignore[assignment]


class ResponseSpec(BaseModel):
    description: str
    model: Any = None
    media_type: Optional[str] = None
    headers: Optional[Mapping[str, Union[Header, str]]] = None
    examples: Optional[Mapping[str, Union[Example, Any]]] = None


class JsonResponseSpec(ResponseSpec):
    model: Any = Field(default={})
    media_type: str = "application/json"


class FileResponseSpec(ResponseSpec):
    model: Any = Field(default=bytes)
    media_type: str = "application/octetstream"


class HTMLResponseSpec(ResponseSpec):
    model: Any = Field(default=str)
    media_type: str = "text/html"



import typing
from typing import List, Type

from pydantic import BaseModel
from pydantic import ValidationError as PydanticValidationError
from pydantic import create_model
from pydantic.error_wrappers import ErrorWrapper
from starlette.exceptions import HTTPException as StarletteHTTPException
from starlette.status import HTTP_422_UNPROCESSABLE_ENTITY


class xpressoError(Exception):
    ...


# This should be upstreamed
class HTTPException(StarletteHTTPException, xpressoError):
    def __init__(
        self,
        status_code: int,
        detail: typing.Any = None,
        headers: typing.Optional[typing.Dict[str, typing.Any]] = None,
    ) -> None:
        super().__init__(status_code=status_code, detail=detail)
        self.headers = headers


_RequestErrorModel: Type[BaseModel] = create_model("Request")  # type: ignore


class RequestValidationError(PydanticValidationError, xpressoError):
    raw_errors: List[ErrorWrapper]

    def __init__(
        self,
        errors: typing.Optional[typing.List[ErrorWrapper]] = None,
        status_code: int = HTTP_422_UNPROCESSABLE_ENTITY,
    ) -> None:
        super().__init__(errors or [], _RequestErrorModel)
        self.status_code = status_code
